﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using HeuristicLab.Grammars;
using Attribute = HeuristicLab.Grammars.Attribute;

namespace CodeGenerator {
  public class RandomSearchCodeGen {

    private string usings = @"
using System.Collections.Generic;
using System.Linq;
using System;
";

    private string solverTemplate = @"
namespace ?PROBLEMNAME? {
  public sealed class ?IDENT?Solver {

    public static void Main(string[] args) {
      var solver = new ?IDENT?Solver();
      solver.Start();
    }

    public ?IDENT?Solver() { 
      Initialize();
    }    

    private void Initialize() {
      ?INITCODE?
    }

    private bool IsBetter(double a, double b) { 
      return ?MAXIMIZATION? ? a > b : a < b;
    }


    private Random seedRandom;
    private int currentSeed;
    private void Start() {
      seedRandom = new Random();
      var bestF = ?MAXIMIZATION? ? double.NegativeInfinity : double.PositiveInfinity;
      int n = 0;
      var sw = new System.Diagnostics.Stopwatch();
      sw.Start();
      while (true) {

        // must make sure that calling the start-symbol multiple times in the fitness function always leads to the same path through the grammar
        // so we use a PRNG for generating seeds for a separate PRNG that is reset each time the start symbol is called
        currentSeed = seedRandom.Next();
        var f = Calculate();

        n++;
        if (IsBetter(f, bestF)) bestF = f;
        if (n % 100 == 0) {
          sw.Stop();
          Console.WriteLine(""{0}\t{1}\t{2}\t({3:0.00} sols/ms)"", n, bestF, f, 100.0 / sw.ElapsedMilliseconds);
          sw.Reset();
          sw.Start();
        }
      }
    }

    public double Calculate() {
      ?FITNESSFUNCTION?
    }

    ?ADDITIONALCODE?

    ?INTERPRETERSOURCE?

    ?CONSTRAINTSSOURCE?
  }
}";


    /// <summary>
    /// Generates the source code for a brute force searcher that can be compiled with a C# compiler
    /// </summary>
    /// <param name="ast">An abstract syntax tree for a GPDL file</param>
    public void Generate(GPDefNode ast) {
      var problemSourceCode = new StringBuilder();
      problemSourceCode.AppendLine(usings);

      GenerateProblem(ast, problemSourceCode);

      problemSourceCode.Replace("?PROBLEMNAME?", ast.Name);

      // write the source file to disk
      using (var stream = new StreamWriter(ast.Name + ".cs")) {
        stream.WriteLine(problemSourceCode.ToString());
      }
    }

    private void GenerateProblem(GPDefNode ast, StringBuilder problemSourceCode) {
      var grammar = CreateGrammarFromAst(ast);
      var problemClassCode =
        solverTemplate
          .Replace("?MAXIMIZATION?", ast.FitnessFunctionNode.Maximization.ToString().ToLowerInvariant())
          .Replace("?IDENT?", ast.Name)
          .Replace("?FITNESSFUNCTION?", ast.FitnessFunctionNode.SrcCode)
          .Replace("?INTERPRETERSOURCE?", GenerateInterpreterSource(grammar))
          .Replace("?INITCODE?", ast.InitCodeNode.SrcCode)
          .Replace("?ADDITIONALCODE?", ast.ClassCodeNode.SrcCode)
          .Replace("?CONSTRAINTSSOURCE?", GenerateConstraintMethods(ast.Terminals))
          ;

      problemSourceCode.AppendLine(problemClassCode).AppendLine();
    }

    private AttributedGrammar CreateGrammarFromAst(GPDefNode ast) {
      string startSymbolName = ast.Rules.First().NtSymbol;
      var startSymbolNode = ast.NonTerminals.Single(nt => nt.Ident == startSymbolName);
      // create startSymbol
      var g = new AttributedGrammar(new Symbol(startSymbolName, ParseSymbolAttributes(startSymbolNode.FormalParameters)));
      foreach (var rule in ast.Rules) {
        // create nt-symbol
        var ntSymbolName = rule.NtSymbol;
        var ntSymbolNode = ast.NonTerminals.Single(nt => nt.Ident == ntSymbolName);
        var attributes = ParseSymbolAttributes(ntSymbolNode.FormalParameters);
        var ntSymbol = new Symbol(ntSymbolName, attributes);
        foreach (var alt in GetAlternatives(rule.Alternatives)) {
          g.AddProductionRule(ntSymbol, alt);
        }
        // local initialization code
        if (!string.IsNullOrEmpty(rule.LocalCode)) g.AddLocalDefinitions(ntSymbol, rule.LocalCode);
      }
      return g;
    }

    private IEnumerable<IAttribute> ParseSymbolAttributes(string formalParameters) {
      return (from fieldDef in Util.ExtractParameters(formalParameters)
              select new Attribute(fieldDef.Identifier, fieldDef.Type, AttributeType.Parse(fieldDef.RefOrOut))).
        ToList();
    }

    private IEnumerable<Sequence> GetAlternatives(AlternativesNode altNode) {
      foreach (var alt in altNode.Alternatives) {
        yield return GetSequence(alt.Sequence);
      }
    }

    private Sequence GetSequence(IEnumerable<RuleExprNode> sequence) {
      Debug.Assert(sequence.All(s => s is CallSymbolNode || s is RuleActionNode));
      var l = new List<ISymbol>();
      foreach (var node in sequence) {
        var callSymbolNode = node as CallSymbolNode;
        var actionNode = node as RuleActionNode;
        if (callSymbolNode != null) {
          l.Add(new Symbol(callSymbolNode.Ident, ParseSymbolAttributes(callSymbolNode.ActualParameter)));
        } else if (actionNode != null) {
          l.Add(new SemanticSymbol("SEM", actionNode.SrcCode));
        }
      }
      return new Sequence(l);
    }

    // produces helper methods for the attributes of all terminal nodes
    private string GenerateConstraintMethods(List<SymbolNode> symbols) {
      var sb = new StringBuilder();
      var terminals = symbols.OfType<TerminalNode>();
      foreach (var t in terminals) {
        sb.AppendLine(GenerateConstraintMethods(t));
      }
      return sb.ToString();
    }

    // generates helper methods for the attributes of a given terminal node
    private string GenerateConstraintMethods(TerminalNode t) {
      var sb = new StringBuilder();
      foreach (var c in t.Constraints) {
        var fieldType = t.FieldDefinitions.First(d => d.Identifier == c.Ident).Type;
        if (c.Type == ConstraintNodeType.Range) {
          sb.AppendFormat("public {0} GetMax{1}_{2}() {{ return {3}; }}", fieldType, t.Ident, c.Ident, c.RangeMaxExpression).AppendLine();
          sb.AppendFormat("public {0} GetMin{1}_{2}() {{ return {3}; }}", fieldType, t.Ident, c.Ident, c.RangeMinExpression).AppendLine();
        } else if (c.Type == ConstraintNodeType.Set) {
          sb.AppendFormat("public IEnumerable<{0}> GetAllowed{1}_{2}() {{ return {3} }}", fieldType, t.Ident, c.Ident, c.SetExpression).AppendLine();
        }
      }
      return sb.ToString();
    }

    private string GenerateInterpreterSource(AttributedGrammar grammar) {
      var sb = new StringBuilder();

      // generate methods for all nonterminals and terminals using the grammar instance
      foreach (var s in grammar.NonTerminalSymbols) {
        sb.AppendLine(GenerateInterpreterMethod(grammar, s));
      }
      foreach (var s in grammar.TerminalSymbols) {
        sb.AppendLine(GenerateTerminalInterpreterMethod(s));
      }
      return sb.ToString();
    }

    private string GenerateTerminalInterpreterMethod(ISymbol s) {
      var sb = new StringBuilder();
      // if the terminal symbol has attributes then we must create values for these attributes
      if (!s.Attributes.Any())
        sb.AppendFormat("private void {0}(Random random) {{", s.Name);
      else
        sb.AppendFormat("private void {0}(Random random, {1}) {{", s.Name, s.GetAttributeString());

      // each field must match a formal parameter, assign a value for each parameter
      foreach (var element in s.Attributes) {
        // only works for "IN SET"-Constraints
        sb.AppendFormat("{{ var tmp = GetAllowed{1}_{0}().ToArray(); {0} = tmp[random.Next(tmp.Length)]; }} ", element.Name, s.Name);
      }
      sb.AppendLine("}");
      return sb.ToString();
    }

    private string GenerateInterpreterMethod(AttributedGrammar g, ISymbol s) {
      var sb = new StringBuilder();
      // if this is the start symbol we additionally have to create the method which can be called from the fitness function

      if (g.StartSymbol.Equals(s)) {
        if (!s.Attributes.Any())
          sb.AppendFormat("private void {0}() {{", s.Name);
        else
          sb.AppendFormat("private void {0}({1}) {{", s.Name, s.GetAttributeString());

        // get formal parameters of start symbol
        var attr = g.StartSymbol.Attributes;

        // actual parameter are the same as formalparameter only without type identifier
        string actualParameter;
        if (attr.Any())
          actualParameter = attr.Skip(1).Aggregate(attr.First().AttributeType + " " + attr.First().Name, (str, a) => str + ", " + a.AttributeType + " " + a.Name);
        else
          actualParameter = string.Empty;

        sb.AppendFormat("{0}(new Random(currentSeed), {1});", g.StartSymbol.Name, actualParameter).AppendLine();
        sb.AppendLine("}");
      }

      if (!s.Attributes.Any())
        sb.AppendFormat("private void {0}(Random random) {{", s.Name);
      else
        sb.AppendFormat("private void {0}(Random random, {1}) {{", s.Name, s.GetAttributeString());

      // generate local definitions
      sb.AppendLine(g.GetLocalDefinitions(s));


      var altsWithSemActions = g.GetAlternativesWithSemanticActions(s).ToArray();
      var terminalAlts = altsWithSemActions.Where(alt => alt.Count(g.IsNonTerminal) == 0);
      var nonTerminalAlts = altsWithSemActions.Where(alt => alt.Count(g.IsNonTerminal) > 0);
      bool hasTerminalAlts = terminalAlts.Count() > 0;
      bool hasNonTerminalAlts = nonTerminalAlts.Count() > 0;

      if (altsWithSemActions.Length > 1) {
        // here we need to bias the selection of alternatives (non-terminal vs terminal alternatives) to make sure that
        // terminals are selected with a certain probability to make sure that:
        // 1) we don't create the same small trees all the time (terminals have high probability to be selected)
        // 2) we don't create very big trees by recursing to deep (leads to stack-overflow) (terminals have a low probability to be selected)
        // so we first decide if we want to generate a terminal or non-terminal (50%, 50%) and then choose a symbol in the class randomly
        // the probability of choosing terminals should depend on the depth of the tree (small likelihood to choose terminals for small depths, large likelihood for large depths)
        if (hasTerminalAlts && hasNonTerminalAlts) {
          sb.AppendLine("if(random.NextDouble() < 0.5) {");
          // terminals
          sb.AppendLine("// terminals ");
          GenerateSwitchStatement(sb, terminalAlts);
          sb.AppendLine("} else {");
          // non-terminals
          sb.AppendLine("// non-terminals ");
          GenerateSwitchStatement(sb, nonTerminalAlts);
          sb.AppendLine("}");
        } else if (hasTerminalAlts) {
          sb.AppendLine("// terminals ");
          GenerateSwitchStatement(sb, terminalAlts);
        } else if (hasNonTerminalAlts) {
          sb.AppendLine("// non-terminals ");
          GenerateSwitchStatement(sb, nonTerminalAlts);
        }
      } else {
        foreach (var altSymb in altsWithSemActions.Single()) {
          sb.AppendLine(GenerateSourceForAction(altSymb));
        }
      }
      sb.AppendLine("}");
      return sb.ToString();
    }

    private void GenerateSwitchStatement(StringBuilder sb, IEnumerable<Sequence> alts) {
      sb.AppendFormat("switch(random.Next({0})) {{", alts.Count());
      // generate a case for each alternative
      int i = 0;
      foreach (var alt in alts) {
        sb.AppendFormat("case {0}: {{ ", i).AppendLine();

        // this only works for alternatives with a single non-terminal symbol (ignoring semantic symbols) so far!
        // a way to handle this is through grammar transformation (the examplary grammars all have the correct from)
        Debug.Assert(alt.Count(symb => !(symb is SemanticSymbol)) == 1);
        foreach (var altSymb in alt) {
          sb.AppendLine(GenerateSourceForAction(altSymb));
        }
        i++;
        sb.AppendLine("break;").AppendLine("}");
      }
      sb.AppendLine("default: throw new System.InvalidOperationException();").AppendLine("}");

    }

    // helper for generating calls to other symbol methods
    private string GenerateSourceForAction(ISymbol s) {
      var action = s as SemanticSymbol;
      if (action != null) {
        return action.Code + ";";
      } else {
        if (!s.Attributes.Any())
          return string.Format("{0}(random);", s.Name);
        else
          return string.Format("{0}(random, {1});", s.Name, s.GetAttributeString());
      }
    }
  }
}
